@@ -1,10 +1,10 @@
-/* Copyright (C) 2002, 2003, 2004, 2005, 2007, 2008, 2009
+/* Copyright (C) 2002, 2003, 2004, 2005, 2007, 2008, 2009, 2010, 2011
    Free Software Foundation, Inc.
    Contributed by Andy Vaught
    Namelist input contributed by Paul Thomas
    F2003 I/O support contributed by Jerry DeLisle
 
-This file is part of the GNU Fortran 95 runtime library (libgfortran).
+This file is part of the GNU Fortran runtime library (libgfortran).
 
 Libgfortran is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
@@ -27,6 +27,8 @@ see the files COPYING3 and COPYING.RUNTIME respectively.  If not, see
 
 
 #include "io.h"
+#include "fbuf.h"
+#include "unix.h"
 #include <string.h>
 #include <stdlib.h>
 #include <ctype.h>
@@ -108,7 +110,7 @@ free_saved (st_parameter_dt *dtp)
   if (dtp->u.p.saved_string == NULL)
     return;
 
-  free_mem (dtp->u.p.saved_string);
+  free (dtp->u.p.saved_string);
 
   dtp->u.p.saved_string = NULL;
   dtp->u.p.saved_used = 0;
@@ -126,24 +128,23 @@ free_line (st_parameter_dt *dtp)
   if (dtp->u.p.line_buffer == NULL)
     return;
 
-  free_mem (dtp->u.p.line_buffer);
+  free (dtp->u.p.line_buffer);
   dtp->u.p.line_buffer = NULL;
 }
 
 
-static char
+static int
 next_char (st_parameter_dt *dtp)
 {
   ssize_t length;
   gfc_offset record;
-  char c;
-  int cc;
+  int c;
 
-  if (dtp->u.p.last_char != '\0')
+  if (dtp->u.p.last_char != EOF - 1)
     {
       dtp->u.p.at_eol = 0;
       c = dtp->u.p.last_char;
-      dtp->u.p.last_char = '\0';
+      dtp->u.p.last_char = EOF - 1;
       goto done;
     }
 
@@ -170,7 +171,7 @@ next_char (st_parameter_dt *dtp)
   if (is_array_io (dtp))
     {
       if (dtp->u.p.at_eof)
-	longjmp (*dtp->u.p.eof_jump, 1);
+	return EOF;
 
       /* Check for "end-of-record" condition.  */
       if (dtp->u.p.current_unit->bytes_left == 0)
@@ -190,7 +191,7 @@ next_char (st_parameter_dt *dtp)
 
 	  record *= dtp->u.p.current_unit->recl;
 	  if (sseek (dtp->u.p.current_unit->s, record, SEEK_SET) < 0)
-	    longjmp (*dtp->u.p.eof_jump, 1);
+	    return EOF;
 
 	  dtp->u.p.current_unit->bytes_left = dtp->u.p.current_unit->recl;
 	  goto done;
@@ -201,7 +202,9 @@ next_char (st_parameter_dt *dtp)
 
   if (is_internal_unit (dtp))
     {
-      length = sread (dtp->u.p.current_unit->s, &c, 1);
+      char cc;
+      length = sread (dtp->u.p.current_unit->s, &cc, 1);
+      c = cc;
       if (length < 0)
 	{
 	  generate_error (&dtp->common, LIBERROR_OS, NULL);
@@ -221,7 +224,7 @@ next_char (st_parameter_dt *dtp)
       else
 	{
 	  if (dtp->u.p.at_eof) 
-	    longjmp (*dtp->u.p.eof_jump, 1);
+	    return EOF;
 	  if (length == 0)
 	    {
 	      c = '\n';
@@ -231,23 +234,12 @@ next_char (st_parameter_dt *dtp)
     }
   else
     {
-      cc = fbuf_getc (dtp->u.p.current_unit);
-
-      if (cc == EOF)
-	{
-	  if (dtp->u.p.current_unit->endfile == AT_ENDFILE)
-	    longjmp (*dtp->u.p.eof_jump, 1);
-	  dtp->u.p.current_unit->endfile = AT_ENDFILE;
-	  c = '\n';
-	}
-      else
-	c = (char) cc;
-      if (is_stream_io (dtp) && cc != EOF)
+      c = fbuf_getc (dtp->u.p.current_unit);
+      if (c != EOF && is_stream_io (dtp))
 	dtp->u.p.current_unit->strm_pos++;
-
     }
 done:
-  dtp->u.p.at_eol = (c == '\n' || c == '\r');
+  dtp->u.p.at_eol = (c == '\n' || c == '\r' || c == EOF);
   return c;
 }
 
@@ -255,7 +247,7 @@ done:
 /* Push a character back onto the input.  */
 
 static void
-unget_char (st_parameter_dt *dtp, char c)
+unget_char (st_parameter_dt *dtp, int c)
 {
   dtp->u.p.last_char = c;
 }
@@ -264,33 +256,35 @@ unget_char (st_parameter_dt *dtp, char c)
 /* Skip over spaces in the input.  Returns the nonspace character that
    terminated the eating and also places it back on the input.  */
 
-static char
+static int
 eat_spaces (st_parameter_dt *dtp)
 {
-  char c;
+  int c;
 
   do
-    {
-      c = next_char (dtp);
-    }
-  while (c == ' ' || c == '\t');
+    c = next_char (dtp);
+  while (c != EOF && (c == ' ' || c == '\t'));
 
   unget_char (dtp, c);
   return c;
 }
 
 
-/* This function reads characters through to the end of the current line and
-   just ignores them.  */
+/* This function reads characters through to the end of the current
+   line and just ignores them.  Returns 0 for success and LIBERROR_END
+   if it hit EOF.  */
 
-static void
+static int
 eat_line (st_parameter_dt *dtp)
 {
-  char c;
+  int c;
 
   do
     c = next_char (dtp);
-  while (c != '\n');
+  while (c != EOF && c != '\n');
+  if (c == EOF)
+    return LIBERROR_END;
+  return 0;
 }
 
 
@@ -303,17 +297,21 @@ eat_line (st_parameter_dt *dtp)
    separator, we stop reading.  If there are more input items, we
    continue reading the separator with finish_separator() which takes
    care of the fact that we may or may not have seen a comma as part
-   of the separator.  */
+   of the separator. 
 
-static void
+   Returns 0 for success, and non-zero error code otherwise.  */
+
+static int
 eat_separator (st_parameter_dt *dtp)
 {
-  char c, n;
+  int c, n;
+  int err = 0;
 
   eat_spaces (dtp);
   dtp->u.p.comma_flag = 0;
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    return LIBERROR_END;
   switch (c)
     {
     case ',':
@@ -334,7 +332,8 @@ eat_separator (st_parameter_dt *dtp)
 
     case '\r':
       dtp->u.p.at_eol = 1;
-      n = next_char(dtp);
+      if ((n = next_char(dtp)) == EOF)
+	return LIBERROR_END;
       if (n != '\n')
 	{
 	  unget_char (dtp, n);
@@ -347,16 +346,14 @@ eat_separator (st_parameter_dt *dtp)
 	{
 	  do
 	    {
-	      c = next_char (dtp);
+	      if ((c = next_char (dtp)) == EOF)
+		  return LIBERROR_END;
 	      if (c == '!')
 		{
-		  eat_line (dtp);
-		  c = next_char (dtp);
-		  if (c == '!')
-		    {
-		      eat_line (dtp);
-		      c = next_char (dtp);
-		    }
+		  err = eat_line (dtp);
+		  if (err)
+		    return err;
+		  c = '\n';
 		}
 	    }
 	  while (c == '\n' || c == '\r' || c == ' ' || c == '\t');
@@ -367,9 +364,9 @@ eat_separator (st_parameter_dt *dtp)
     case '!':
       if (dtp->u.p.namelist_mode)
 	{			/* Eat a namelist comment.  */
-	  do
-	    c = next_char (dtp);
-	  while (c != '\n');
+	  err = eat_line (dtp);
+	  if (err)
+	    return err;
 
 	  break;
 	}
@@ -380,22 +377,26 @@ eat_separator (st_parameter_dt *dtp)
       unget_char (dtp, c);
       break;
     }
+  return err;
 }
 
 
 /* Finish processing a separator that was interrupted by a newline.
    If we're here, then another data item is present, so we finish what
-   we started on the previous line.  */
+   we started on the previous line.  Return 0 on success, error code
+   on failure.  */
 
-static void
+static int
 finish_separator (st_parameter_dt *dtp)
 {
-  char c;
+  int c;
+  int err;
 
  restart:
   eat_spaces (dtp);
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    return LIBERROR_END;
   switch (c)
     {
     case ',':
@@ -403,7 +404,8 @@ finish_separator (st_parameter_dt *dtp)
 	unget_char (dtp, c);
       else
 	{
-	  c = eat_spaces (dtp);
+	  if ((c = eat_spaces (dtp)) == EOF)
+	    return LIBERROR_END;
 	  if (c == '\n' || c == '\r')
 	    goto restart;
 	}
@@ -413,7 +415,7 @@ finish_separator (st_parameter_dt *dtp)
     case '/':
       dtp->u.p.input_complete = 1;
       if (!dtp->u.p.namelist_mode)
-	return;
+	return err;
       break;
 
     case '\n':
@@ -423,10 +425,9 @@ finish_separator (st_parameter_dt *dtp)
     case '!':
       if (dtp->u.p.namelist_mode)
 	{
-	  do
-	    c = next_char (dtp);
-	  while (c != '\n');
-
+	  err = eat_line (dtp);
+	  if (err)
+	    return err;
 	  goto restart;
 	}
 
@@ -434,6 +435,7 @@ finish_separator (st_parameter_dt *dtp)
       unget_char (dtp, c);
       break;
     }
+  return err;
 }
 
 
@@ -533,10 +535,11 @@ convert_integer (st_parameter_dt *dtp, int length, int negative)
 static int
 parse_repeat (st_parameter_dt *dtp)
 {
-  char c, message[100];
-  int repeat;
+  char message[100];
+  int c, repeat;
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad_repeat;
   switch (c)
     {
     CASE_DIGITS:
@@ -597,8 +600,14 @@ parse_repeat (st_parameter_dt *dtp)
 
  bad_repeat:
 
-  eat_line (dtp);
   free_saved (dtp);
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return 1;
+    }
+  else
+    eat_line (dtp);
   sprintf (message, "Bad repeat count in item %d of list input",
 	   dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -629,8 +638,8 @@ l_push_char (st_parameter_dt *dtp, char c)
 static void
 read_logical (st_parameter_dt *dtp, int length)
 {
-  char c, message[100];
-  int i, v;
+  char message[100];
+  int c, i, v;
 
   if (parse_repeat (dtp))
     return;
@@ -644,7 +653,7 @@ read_logical (st_parameter_dt *dtp, int length)
       c = next_char (dtp);
       l_push_char (dtp, c);
 
-      if (!is_separator(c))
+      if (!is_separator(c) && c != EOF)
 	goto possible_name;
 
       unget_char (dtp, c);
@@ -654,7 +663,7 @@ read_logical (st_parameter_dt *dtp, int length)
       c = next_char (dtp);
       l_push_char (dtp, c);
 
-      if (!is_separator(c))
+      if (!is_separator(c) && c != EOF)
 	goto possible_name;
 
       unget_char (dtp, c);
@@ -693,10 +702,8 @@ read_logical (st_parameter_dt *dtp, int length)
 
   /* Eat trailing garbage.  */
   do
-    {
-      c = next_char (dtp);
-    }
-  while (!is_separator (c));
+    c = next_char (dtp);
+  while (c != EOF && !is_separator (c));
 
   unget_char (dtp, c);
   eat_separator (dtp);
@@ -744,8 +751,14 @@ read_logical (st_parameter_dt *dtp, int length)
   if (nml_bad_return (dtp, c))
     return;
 
-  eat_line (dtp);
   free_saved (dtp);
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return;
+    }
+  else if (c != '\n')
+    eat_line (dtp);
   sprintf (message, "Bad logical value while reading item %d",
 	      dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -769,8 +782,8 @@ read_logical (st_parameter_dt *dtp, int length)
 static void
 read_integer (st_parameter_dt *dtp, int length)
 {
-  char c, message[100];
-  int negative;
+  char message[100];
+  int c, negative;
 
   negative = 0;
 
@@ -782,7 +795,8 @@ read_integer (st_parameter_dt *dtp, int length)
       /* Fall through...  */
 
     case '+':
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto bad_integer;
       goto get_integer;
 
     CASE_SEPARATORS:		/* Single null.  */
@@ -814,6 +828,7 @@ read_integer (st_parameter_dt *dtp, int length)
 	  goto repeat;
 
 	CASE_SEPARATORS:	/* Not a repeat count.  */
+	case EOF:
 	  goto done;
 
 	default:
@@ -827,7 +842,8 @@ read_integer (st_parameter_dt *dtp, int length)
 
   /* Get the real integer.  */
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad_integer;
   switch (c)
     {
     CASE_DIGITS:
@@ -861,7 +877,7 @@ read_integer (st_parameter_dt *dtp, int length)
 	  push_char (dtp, c);
 	  break;
 
-	CASE_SEPARATORS:
+	CASE_SEPARATORS:	  
 	  goto done;
 
 	default:
@@ -873,9 +889,15 @@ read_integer (st_parameter_dt *dtp, int length)
 
   if (nml_bad_return (dtp, c))
     return;
-  
-  eat_line (dtp);
-  free_saved (dtp);
+
+  free_saved (dtp);  
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return;
+    }
+  else if (c != '\n')
+    eat_line (dtp);
   sprintf (message, "Bad integer for item %d in list input",
 	      dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -903,11 +925,13 @@ read_integer (st_parameter_dt *dtp, int length)
 static void
 read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 {
-  char c, quote, message[100];
+  char quote, message[100];
+  int c;
 
   quote = ' ';			/* Space means no quote character.  */
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto eof;
   switch (c)
     {
     CASE_DIGITS:
@@ -939,7 +963,8 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 
   for (;;)
     {
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto eof;
       switch (c)
 	{
 	CASE_DIGITS:
@@ -966,7 +991,8 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 
   /* Now get the real string.  */
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto eof;
   switch (c)
     {
     CASE_SEPARATORS:
@@ -987,7 +1013,8 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
  get_string:
   for (;;)
     {
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto done_eof;
       switch (c)
 	{
 	case '"':
@@ -1001,7 +1028,8 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 	  /* See if we have a doubled quote character or the end of
 	     the string.  */
 
-	  c = next_char (dtp);
+	  if ((c = next_char (dtp)) == EOF)
+	    goto eof;
 	  if (c == quote)
 	    {
 	      push_char (dtp, quote);
@@ -1032,7 +1060,8 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
      invalid.  */
  done:
   c = next_char (dtp);
-  if (is_separator (c) || c == '!')
+ done_eof:
+  if (is_separator (c) || c == '!' || c == EOF)
     {
       unget_char (dtp, c);
       eat_separator (dtp);
@@ -1046,6 +1075,11 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 		  dtp->u.p.item_count);
       generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
     }
+  return;
+
+ eof:
+  free_saved (dtp);
+  hit_eof (dtp);
 }
 
 
@@ -1055,14 +1089,17 @@ read_character (st_parameter_dt *dtp, int length __attribute__ ((unused)))
 static int
 parse_real (st_parameter_dt *dtp, void *buffer, int length)
 {
-  char c, message[100];
-  int m, seen_dp;
+  char message[100];
+  int c, m, seen_dp;
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad;
+    
   if (c == '-' || c == '+')
     {
       push_char (dtp, c);
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto bad;
     }
 
   if (c == ',' && dtp->u.p.current_unit->decimal_status == DECIMAL_COMMA)
@@ -1082,7 +1119,8 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
 
   for (;;)
     {
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto bad;
       if (c == ',' && dtp->u.p.current_unit->decimal_status == DECIMAL_COMMA)
 	c = '.';
       switch (c)
@@ -1110,11 +1148,11 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
 	case '+':
 	  push_char (dtp, 'e');
 	  push_char (dtp, c);
-	  c = next_char (dtp);
+	  if ((c = next_char (dtp)) == EOF)
+	    goto bad;
 	  goto exp2;
 
 	CASE_SEPARATORS:
-	  unget_char (dtp, c);
 	  goto done;
 
 	default:
@@ -1123,7 +1161,8 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
     }
 
  exp1:
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad;
   if (c != '-' && c != '+')
     push_char (dtp, '+');
   else
@@ -1140,7 +1179,8 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
 
   for (;;)
     {
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto bad;
       switch (c)
 	{
 	CASE_DIGITS:
@@ -1197,6 +1237,18 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
       push_char (dtp, 'n');
       push_char (dtp, 'a');
       push_char (dtp, 'n');
+      
+      /* Match "NAN(alphanum)".  */
+      if (c == '(')
+	{
+	  for ( ; c != ')'; c = next_char (dtp))
+	    if (is_separator (c))
+	      goto bad;
+
+	  c = next_char (dtp);
+	  if (is_separator (c))
+	    unget_char (dtp, c);
+	}
       goto done;
     }
 
@@ -1205,8 +1257,14 @@ parse_real (st_parameter_dt *dtp, void *buffer, int length)
   if (nml_bad_return (dtp, c))
     return 0;
 
-  eat_line (dtp);
   free_saved (dtp);
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return 1;
+    }
+  else if (c != '\n')
+    eat_line (dtp);
   sprintf (message, "Bad floating point number for item %d",
 	      dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -1222,7 +1280,7 @@ static void
 read_complex (st_parameter_dt *dtp, void * dest, int kind, size_t size)
 {
   char message[100];
-  char c;
+  int c;
 
   if (parse_repeat (dtp))
     return;
@@ -1242,10 +1300,6 @@ read_complex (st_parameter_dt *dtp, void * dest, int kind, size_t size)
       goto bad_complex;
     }
 
-  eat_spaces (dtp);
-  if (parse_real (dtp, dest, kind))
-    return;
-
 eol_1:
   eat_spaces (dtp);
   c = next_char (dtp);
@@ -1254,9 +1308,8 @@ eol_1:
   else
     unget_char (dtp, c);
 
-  if (next_char (dtp)
-      !=  (dtp->u.p.current_unit->decimal_status == DECIMAL_POINT ? ',' : ';'))
-    goto bad_complex;
+  if (parse_real (dtp, dest, kind))
+    return;
 
 eol_2:
   eat_spaces (dtp);
@@ -1266,10 +1319,29 @@ eol_2:
   else
     unget_char (dtp, c);
 
+  if (next_char (dtp)
+      !=  (dtp->u.p.current_unit->decimal_status == DECIMAL_POINT ? ',' : ';'))
+    goto bad_complex;
+
+eol_3:
+  eat_spaces (dtp);
+  c = next_char (dtp);
+  if (c == '\n' || c== '\r')
+    goto eol_3;
+  else
+    unget_char (dtp, c);
+
   if (parse_real (dtp, dest + size / 2, kind))
     return;
-
+    
+eol_4:
   eat_spaces (dtp);
+  c = next_char (dtp);
+  if (c == '\n' || c== '\r')
+    goto eol_4;
+  else
+    unget_char (dtp, c);
+
   if (next_char (dtp) != ')')
     goto bad_complex;
 
@@ -1289,8 +1361,14 @@ eol_2:
   if (nml_bad_return (dtp, c))
     return;
 
-  eat_line (dtp);
   free_saved (dtp);
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return;
+    }
+  else if (c != '\n')   
+    eat_line (dtp);
   sprintf (message, "Bad complex value in item %d of list input",
 	      dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -1302,7 +1380,8 @@ eol_2:
 static void
 read_real (st_parameter_dt *dtp, void * dest, int length)
 {
-  char c, message[100];
+  char message[100];
+  int c;
   int seen_dp;
   int is_inf;
 
@@ -1395,7 +1474,8 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
 
   /* Now get the number itself.  */
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad_real;
   if (is_separator (c))
     {				/* Repeated null value.  */
       unget_char (dtp, c);
@@ -1409,7 +1489,8 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
     {
     got_sign:
       push_char (dtp, c);
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	goto bad_real;
     }
 
   if (c == ',' && dtp->u.p.current_unit->decimal_status == DECIMAL_COMMA)
@@ -1446,6 +1527,7 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
 	  break;
 
 	CASE_SEPARATORS:
+	case EOF:
 	  goto done;
 
 	case '.':
@@ -1477,7 +1559,8 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
  exp1:
   push_char (dtp, 'e');
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    goto bad_real;
   if (c != '+' && c != '-')
     push_char (dtp, '+');
   else
@@ -1502,6 +1585,7 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
 	  break;
 
 	CASE_SEPARATORS:
+	case EOF:
 	  goto done;
 
 	default:
@@ -1574,6 +1658,20 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
 	goto unwind;
       c = next_char (dtp);
       l_push_char (dtp, c);
+
+      /* Match NAN(alphanum).  */
+      if (c == '(')
+	{
+	  for (c = next_char (dtp); c != ')'; c = next_char (dtp))
+	    if (is_separator (c))
+	      goto unwind;
+	    else
+	      l_push_char (dtp, c);
+
+	  l_push_char (dtp, ')');
+	  c = next_char (dtp);
+	  l_push_char (dtp, c);
+	}
     }
 
   if (!is_separator (c))
@@ -1584,7 +1682,10 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
       if (c == ' ' || c =='\n' || c == '\r')
 	{
 	  do
-	    c = next_char (dtp);
+	    {
+	      if ((c = next_char (dtp)) == EOF)
+		goto bad_real;
+	    }
 	  while (c == ' ' || c =='\n' || c == '\r');
 
 	  l_push_char (dtp, c);
@@ -1624,8 +1725,15 @@ read_real (st_parameter_dt *dtp, void * dest, int length)
   if (nml_bad_return (dtp, c))
     return;
 
-  eat_line (dtp);
   free_saved (dtp);
+  if (c == EOF)
+    {
+      hit_eof (dtp);
+      return;
+    }
+  else if (c != '\n')
+    eat_line (dtp);
+
   sprintf (message, "Bad real number in item %d of list input",
 	      dtp->u.p.item_count);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, message);
@@ -1640,7 +1748,7 @@ check_type (st_parameter_dt *dtp, bt type, int len)
 {
   char message[100];
 
-  if (dtp->u.p.saved_type != BT_NULL && dtp->u.p.saved_type != type)
+  if (dtp->u.p.saved_type != BT_UNKNOWN && dtp->u.p.saved_type != type)
     {
       sprintf (message, "Read type %s where %s was expected for item %d",
 		  type_name (dtp->u.p.saved_type), type_name (type),
@@ -1650,7 +1758,7 @@ check_type (st_parameter_dt *dtp, bt type, int len)
       return 1;
     }
 
-  if (dtp->u.p.saved_type == BT_NULL || dtp->u.p.saved_type == BT_CHARACTER)
+  if (dtp->u.p.saved_type == BT_UNKNOWN || dtp->u.p.saved_type == BT_CHARACTER)
     return 0;
 
   if (dtp->u.p.saved_length != len)
@@ -1672,29 +1780,16 @@ check_type (st_parameter_dt *dtp, bt type, int len)
    reading, usually in the dtp->u.p.value[] array.  If a repeat count is
    greater than one, we copy the data item multiple times.  */
 
-static void
-list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
+static int
+list_formatted_read_scalar (st_parameter_dt *dtp, bt type, void *p,
 			    int kind, size_t size)
 {
-  char c;
   gfc_char4_t *q;
-  int i, m;
-  jmp_buf eof_jump;
+  int c, i, m;
+  int err = 0;
 
   dtp->u.p.namelist_mode = 0;
 
-  dtp->u.p.eof_jump = &eof_jump;
-  if (setjmp (eof_jump))
-    {
-      generate_error (&dtp->common, LIBERROR_END, NULL);
-    if (!is_internal_unit (dtp))
-      {
-        dtp->u.p.current_unit->endfile = AFTER_ENDFILE;
-        dtp->u.p.current_unit->current_record = 0;
-      }
-      goto cleanup;
-    }
-
   if (dtp->u.p.first_item)
     {
       dtp->u.p.first_item = 0;
@@ -1702,7 +1797,11 @@ list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
       dtp->u.p.repeat_count = 1;
       dtp->u.p.at_eol = 0;
       
-      c = eat_spaces (dtp);
+      if ((c = eat_spaces (dtp)) == EOF)
+	{
+	  err = LIBERROR_END;
+	  goto cleanup;
+	}
       if (is_separator (c))
 	{
 	  /* Found a null value.  */
@@ -1726,16 +1825,13 @@ list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
       if (dtp->u.p.repeat_count > 0)
 	{
 	  if (check_type (dtp, type, kind))
-	    return;
+	    return err;
 	  goto set_value;
 	}
 	
       if (dtp->u.p.input_complete)
 	goto cleanup;
 
-      if (dtp->u.p.input_complete)
-	goto cleanup;
-
       if (dtp->u.p.at_eol)
 	finish_separator (dtp);
       else
@@ -1746,7 +1842,7 @@ list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
 	    finish_separator (dtp);
         }
 
-      dtp->u.p.saved_type = BT_NULL;
+      dtp->u.p.saved_type = BT_UNKNOWN;
       dtp->u.p.repeat_count = 1;
     }
 
@@ -1777,7 +1873,7 @@ list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
       internal_error (&dtp->common, "Bad type for list read");
     }
 
-  if (dtp->u.p.saved_type != BT_CHARACTER && dtp->u.p.saved_type != BT_NULL)
+  if (dtp->u.p.saved_type != BT_CHARACTER && dtp->u.p.saved_type != BT_UNKNOWN)
     dtp->u.p.saved_length = size;
 
   if ((dtp->common.flags & IOPARM_LIBRETURN_MASK) != IOPARM_LIBRETURN_OK)
@@ -1828,15 +1924,20 @@ list_formatted_read_scalar (st_parameter_dt *dtp, volatile bt type, void *p,
 	}
       break;
 
-    case BT_NULL:
+    case BT_UNKNOWN:
       break;
+
+    default:
+      internal_error (&dtp->common, "Bad type for list read");
     }
 
   if (--dtp->u.p.repeat_count <= 0)
     free_saved (dtp);
 
 cleanup:
-  dtp->u.p.eof_jump = NULL;
+  if (err == LIBERROR_END)
+    hit_eof (dtp);
+  return err;
 }
 
 
@@ -1848,6 +1949,7 @@ list_formatted_read (st_parameter_dt *dtp, bt type, void *p, int kind,
   char *tmp;
   size_t stride = type == BT_CHARACTER ?
 		  size * GFC_SIZE_OF_CHAR_KIND(kind) : size;
+  int err;
 
   tmp = (char *) p;
 
@@ -1855,7 +1957,10 @@ list_formatted_read (st_parameter_dt *dtp, bt type, void *p, int kind,
   for (elem = 0; elem < nelems; elem++)
     {
       dtp->u.p.item_count++;
-      list_formatted_read_scalar (dtp, type, tmp + stride*elem, kind, size);
+      err = list_formatted_read_scalar (dtp, type, tmp + stride*elem, 
+					kind, size);
+      if (err)
+	break;
     }
 }
 
@@ -1865,7 +1970,7 @@ list_formatted_read (st_parameter_dt *dtp, bt type, void *p, int kind,
 void
 finish_list_read (st_parameter_dt *dtp)
 {
-  char c;
+  int err;
 
   free_saved (dtp);
 
@@ -1877,18 +1982,9 @@ finish_list_read (st_parameter_dt *dtp)
       return;
     }
 
-  do
-    {
-      c = next_char (dtp);
-    }
-  while (c != '\n');
-
-  if (dtp->u.p.current_unit->endfile != NO_ENDFILE)
-    {
-      generate_error (&dtp->common, LIBERROR_END, NULL);
-      dtp->u.p.current_unit->endfile = AFTER_ENDFILE;
-      dtp->u.p.current_unit->current_record = 0;
-    }
+  err = eat_line (dtp);
+  if (err == LIBERROR_END)
+    hit_eof (dtp);
 }
 
 /*			NAMELIST INPUT
@@ -1925,7 +2021,7 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
   int neg;
   int null_flag;
   int is_array_section, is_char;
-  char c;
+  int c;
 
   is_char = 0;
   is_array_section = 0;
@@ -1940,7 +2036,8 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 
   /* The next character in the stream should be the '('.  */
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    return FAILURE;
 
   /* Process the qualifier, by dimension and triplet.  */
 
@@ -1953,7 +2050,8 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 	  neg = 0;
 
 	  /* Process a potential sign.  */
-	  c = next_char (dtp);
+	  if ((c = next_char (dtp)) == EOF)
+	    return FAILURE;
 	  switch (c)
 	    {
 	    case '-':
@@ -1971,7 +2069,8 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 	  /* Process characters up to the next ':' , ',' or ')'.  */
 	  for (;;)
 	    {
-	      c = next_char (dtp);
+	      if ((c = next_char (dtp)) == EOF)
+		return FAILURE;
 
 	      switch (c)
 		{
@@ -1997,7 +2096,8 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 
 		case ' ': case '\t':
 		  eat_spaces (dtp);
-		  c = next_char (dtp);
+		  if ((c = next_char (dtp) == EOF))
+		    return FAILURE;
 		  break;
 
 		default:
@@ -2077,8 +2177,9 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 
 		  /*  If -std=f95/2003 or an array section is specified,
 		      do not allow excess data to be processed.  */
-                  if (is_array_section == 1
-		      || compile_options.allow_std < GFC_STD_GNU)
+		  if (is_array_section == 1
+		      || !(compile_options.allow_std & GFC_STD_GNU)
+		      || dtp->u.p.ionml->type == BT_DERIVED)
 		    ls[dim].end = ls[dim].start;
 		  else
 		    dtp->u.p.expanded_read = 1;
@@ -2093,18 +2194,18 @@ nml_parse_qualifier (st_parameter_dt *dtp, descriptor_dimension *ad,
 	}
 
       if (is_array_section == 1 && dtp->u.p.expanded_read == 1)
-     	{
+	{
 	  int i;
 	  dtp->u.p.expanded_read = 0;
 	  for (i = 0; i < dim; i++)
 	    ls[i].end = ls[i].start;
-      	}
+	}
 
       /* Check the values of the triplet indices.  */
-      if ((ls[dim].start > (ssize_t)ad[dim].ubound)
-	  || (ls[dim].start < (ssize_t)ad[dim].lbound)
-	  || (ls[dim].end > (ssize_t)ad[dim].ubound)
-	  || (ls[dim].end < (ssize_t)ad[dim].lbound))
+      if ((ls[dim].start > (ssize_t) GFC_DIMENSION_UBOUND(ad[dim]))
+	   || (ls[dim].start < (ssize_t) GFC_DIMENSION_LBOUND(ad[dim]))
+	   || (ls[dim].end > (ssize_t) GFC_DIMENSION_UBOUND(ad[dim]))
+	   || (ls[dim].end < (ssize_t) GFC_DIMENSION_LBOUND(ad[dim])))
 	{
 	  if (is_char)
 	    sprintf (parse_err_msg, "Substring out of range");
@@ -2168,15 +2269,15 @@ nml_touch_nodes (namelist_info * nl)
 	  for (dim=0; dim < nl->var_rank; dim++)
 	    {
 	      nl->ls[dim].step = 1;
-	      nl->ls[dim].end = nl->dim[dim].ubound;
-	      nl->ls[dim].start = nl->dim[dim].lbound;
+	      nl->ls[dim].end = GFC_DESCRIPTOR_UBOUND(nl,dim);
+	      nl->ls[dim].start = GFC_DESCRIPTOR_LBOUND(nl,dim);
 	      nl->ls[dim].idx = nl->ls[dim].start;
 	    }
 	}
       else
 	break;
     }
-  free_mem (ext_name);
+  free (ext_name);
   return;
 }
 
@@ -2199,12 +2300,13 @@ static void
 nml_match_name (st_parameter_dt *dtp, const char *name, index_type len)
 {
   index_type i;
-  char c;
+  int c;
+
   dtp->u.p.nml_read_error = 0;
   for (i = 0; i < len; i++)
     {
       c = next_char (dtp);
-      if (tolower (c) != tolower (name[i]))
+      if (c == EOF || (tolower (c) != tolower (name[i])))
 	{
 	  dtp->u.p.nml_read_error = 1;
 	  break;
@@ -2225,11 +2327,11 @@ nml_query (st_parameter_dt *dtp, char c)
   index_type len;
   char * p;
 #ifdef HAVE_CRLF
-  static const index_type endlen = 3;
+  static const index_type endlen = 2;
   static const char endl[] = "\r\n";
   static const char nmlend[] = "&end\r\n";
 #else
-  static const index_type endlen = 2;
+  static const index_type endlen = 1;
   static const char endl[] = "\n";
   static const char nmlend[] = "&end\n";
 #endif
@@ -2259,12 +2361,12 @@ nml_query (st_parameter_dt *dtp, char c)
 	  /* "&namelist_name\n"  */
 
 	  len = dtp->namelist_name_len;
-	  p = write_block (dtp, len + endlen);
+	  p = write_block (dtp, len - 1 + endlen);
           if (!p)
             goto query_return;
 	  memcpy (p, "&", 1);
 	  memcpy ((char*)(p + 1), dtp->namelist_name, len);
-	  memcpy ((char*)(p + len + 1), &endl, endlen - 1);
+	  memcpy ((char*)(p + len + 1), &endl, endlen);
 	  for (nl = dtp->u.p.ionml; nl; nl = nl->next)
 	    {
 	      /* " var_name\n"  */
@@ -2275,14 +2377,15 @@ nml_query (st_parameter_dt *dtp, char c)
 		goto query_return;
 	      memcpy (p, " ", 1);
 	      memcpy ((char*)(p + 1), nl->var_name, len);
-	      memcpy ((char*)(p + len + 1), &endl, endlen - 1);
+	      memcpy ((char*)(p + len + 1), &endl, endlen);
 	    }
 
 	  /* "&end\n"  */
 
-          p = write_block (dtp, endlen + 3);
+          p = write_block (dtp, endlen + 4);
+	  if (!p)
 	    goto query_return;
-          memcpy (p, &nmlend, endlen + 3);
+          memcpy (p, &nmlend, endlen + 4);
 	}
 
       /* Flush the stream to force immediate output.  */
@@ -2337,20 +2440,20 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
   len = nl->len;
   switch (nl->type)
   {
-    case GFC_DTYPE_INTEGER:
-    case GFC_DTYPE_LOGICAL:
+    case BT_INTEGER:
+    case BT_LOGICAL:
       dlen = len;
       break;
 
-    case GFC_DTYPE_REAL:
+    case BT_REAL:
       dlen = size_from_real_kind (len);
       break;
 
-    case GFC_DTYPE_COMPLEX:
+    case BT_COMPLEX:
       dlen = size_from_complex_kind (len);
       break;
 
-    case GFC_DTYPE_CHARACTER:
+    case BT_CHARACTER:
       dlen = chigh ? (chigh - clow + 1) : nl->string_length;
       break;
 
@@ -2364,8 +2467,9 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
 
       pdata = (void*)(nl->mem_pos + offset);
       for (dim = 0; dim < nl->var_rank; dim++)
-	pdata = (void*)(pdata + (nl->ls[dim].idx - nl->dim[dim].lbound) *
-		 nl->dim[dim].stride * nl->size);
+	pdata = (void*)(pdata + (nl->ls[dim].idx
+				 - GFC_DESCRIPTOR_LBOUND(nl,dim))
+			* GFC_DESCRIPTOR_STRIDE(nl,dim) * nl->size);
 
       /* Reset the error flag and try to read next value, if
 	 dtp->u.p.repeat_count=0  */
@@ -2381,40 +2485,37 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
 	  if (dtp->u.p.input_complete)
 	    return SUCCESS;
 
-	  /* GFC_TYPE_UNKNOWN through for nulls and is detected
-	     after the switch block.  */
-
-	  dtp->u.p.saved_type = GFC_DTYPE_UNKNOWN;
+	  dtp->u.p.saved_type = BT_UNKNOWN;
 	  free_saved (dtp);
 
           switch (nl->type)
 	  {
-	  case GFC_DTYPE_INTEGER:
+	  case BT_INTEGER:
 	      read_integer (dtp, len);
               break;
 
-	  case GFC_DTYPE_LOGICAL:
+	  case BT_LOGICAL:
 	      read_logical (dtp, len);
               break;
 
-	  case GFC_DTYPE_CHARACTER:
+	  case BT_CHARACTER:
 	      read_character (dtp, len);
               break;
 
-	  case GFC_DTYPE_REAL:
+	  case BT_REAL:
 	    /* Need to copy data back from the real location to the temp in order
 	       to handle nml reads into arrays.  */
 	    read_real (dtp, pdata, len);
 	    memcpy (dtp->u.p.value, pdata, dlen);
 	    break;
 
-	  case GFC_DTYPE_COMPLEX:
+	  case BT_COMPLEX:
 	    /* Same as for REAL, copy back to temp.  */
 	    read_complex (dtp, pdata, len, dlen);
 	    memcpy (dtp->u.p.value, pdata, dlen);
 	    break;
 
-	  case GFC_DTYPE_DERIVED:
+	  case BT_DERIVED:
 	    obj_name_len = strlen (nl->var_name) + 1;
 	    obj_name = get_mem (obj_name_len+1);
 	    memcpy (obj_name, nl->var_name, obj_name_len-1);
@@ -2440,18 +2541,18 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
 				  pprev_nl, nml_err_msg, nml_err_msg_size,
 				  clow, chigh) == FAILURE)
 		  {
-		    free_mem (obj_name);
+		    free (obj_name);
 		    return FAILURE;
 		  }
 
 		if (dtp->u.p.input_complete)
 		  {
-		    free_mem (obj_name);
+		    free (obj_name);
 		    return SUCCESS;
 		  }
 	      }
 
-	    free_mem (obj_name);
+	    free (obj_name);
 	    goto incr_idx;
 
           default:
@@ -2474,15 +2575,12 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
 	  return SUCCESS;
 	}
 
-      if (dtp->u.p.saved_type == GFC_DTYPE_UNKNOWN)
+      if (dtp->u.p.saved_type == BT_UNKNOWN)
 	{
 	  dtp->u.p.expanded_read = 0;
 	  goto incr_idx;
 	}
 
-      /* Note the switch from GFC_DTYPE_type to BT_type at this point.
-	 This comes about because the read functions return BT_types.  */
-
       switch (dtp->u.p.saved_type)
       {
 
@@ -2494,7 +2592,19 @@ nml_read_obj (st_parameter_dt *dtp, namelist_info * nl, index_type offset,
 	  break;
 
 	case BT_CHARACTER:
-	  m = (dlen < dtp->u.p.saved_used) ? dlen : dtp->u.p.saved_used;
+	  if (dlen < dtp->u.p.saved_used)
+	    {
+	      if (compile_options.bounds_check)
+		{
+		  snprintf (nml_err_msg, nml_err_msg_size,
+			    "Namelist object '%s' truncated on read.",
+			    nl->var_name);
+		  generate_warning (&dtp->common, nml_err_msg);
+		}
+	      m = dlen;
+	    }
+	  else
+	    m = dtp->u.p.saved_used;
 	  pdata = (void*)( pdata + clow - 1 );
 	  memcpy (pdata, dtp->u.p.saved_string, m);
 	  if (m < dlen)
@@ -2566,7 +2676,7 @@ static try
 nml_get_obj_data (st_parameter_dt *dtp, namelist_info **pprev_nl,
 		  char *nml_err_msg, size_t nml_err_msg_size)
 {
-  char c;
+  int c;
   namelist_info * nl;
   namelist_info * first_nl = NULL;
   namelist_info * root_nl = NULL;
@@ -2587,11 +2697,13 @@ nml_get_obj_data (st_parameter_dt *dtp, namelist_info **pprev_nl,
   if (dtp->u.p.input_complete)
     return SUCCESS;
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    return FAILURE;
   switch (c)
     {
     case '=':
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	return FAILURE;
       if (c != '?')
 	{
 	  sprintf (nml_err_msg, "namelist read: misplaced = sign");
@@ -2638,7 +2750,8 @@ get_name:
     {
       if (!is_separator (c))
 	push_char (dtp, tolower(c));
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	return FAILURE;
     } while (!( c=='=' || c==' ' || c=='\t' || c =='(' || c =='%' ));
 
   unget_char (dtp, c);
@@ -2688,8 +2801,8 @@ get_name:
   for (dim=0; dim < nl->var_rank; dim++)
     {
       nl->ls[dim].step = 1;
-      nl->ls[dim].end = nl->dim[dim].ubound;
-      nl->ls[dim].start = nl->dim[dim].lbound;
+      nl->ls[dim].end = GFC_DESCRIPTOR_UBOUND(nl,dim);
+      nl->ls[dim].start = GFC_DESCRIPTOR_LBOUND(nl,dim);
       nl->ls[dim].idx = nl->ls[dim].start;
     }
 
@@ -2712,7 +2825,8 @@ get_name:
 
       qualifier_flag = 1;
 
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	return FAILURE;
       unget_char (dtp, c);
     }
   else if (nl->var_rank > 0)
@@ -2724,7 +2838,7 @@ get_name:
 
   if (c == '%')
     {
-      if (nl->type != GFC_DTYPE_DERIVED)
+      if (nl->type != BT_DERIVED)
 	{
 	  snprintf (nml_err_msg, nml_err_msg_size,
 		    "Attempt to get derived component for %s", nl->var_name);
@@ -2737,8 +2851,8 @@ get_name:
       root_nl = nl;
 
       component_flag = 1;
-
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	return FAILURE;
       goto get_name;
     }
 
@@ -2748,7 +2862,7 @@ get_name:
   clow = 1;
   chigh = 0;
 
-  if (c == '(' && nl->type == GFC_DTYPE_CHARACTER)
+  if (c == '(' && nl->type == BT_CHARACTER)
     {
       descriptor_dimension chd[1] = { {1, clow, nl->string_length} };
       array_loop_spec ind[1] = { {1, clow, nl->string_length, 1} };
@@ -2774,7 +2888,8 @@ get_name:
 	  goto nml_err_ret;
 	}
 
-      c = next_char (dtp);
+      if ((c = next_char (dtp)) == EOF)
+	return FAILURE;
       unget_char (dtp, c);
     }
 
@@ -2813,7 +2928,8 @@ get_name:
   if (dtp->u.p.input_complete)
     return SUCCESS;
 
-  c = next_char (dtp);
+  if ((c = next_char (dtp)) == EOF)
+    return FAILURE;
 
   if (c != '=')
     {
@@ -2826,7 +2942,7 @@ get_name:
      namelist_info if we have parsed a qualified derived type
      component.  */
 
-  if (nl->type == GFC_DTYPE_DERIVED)
+  if (nl->type == BT_DERIVED)
     nml_touch_nodes (nl);
 
   if (first_nl)
@@ -2858,9 +2974,13 @@ nml_err_ret:
 void
 namelist_read (st_parameter_dt *dtp)
 {
-  char c;
-  jmp_buf eof_jump;
+  int c;
   char nml_err_msg[200];
+
+  /* Initialize the error string buffer just in case we get an unexpected fail
+     somewhere and end up at nml_err_ret.  */
+  strcpy (nml_err_msg, "Internal namelist read error");
+
   /* Pointer to the previously read object, in case attempt is made to read
      new object name.  Should this fail, error message can give previous
      name.  */
@@ -2870,20 +2990,13 @@ namelist_read (st_parameter_dt *dtp)
   dtp->u.p.input_complete = 0;
   dtp->u.p.expanded_read = 0;
 
-  dtp->u.p.eof_jump = &eof_jump;
-  if (setjmp (eof_jump))
-    {
-      dtp->u.p.eof_jump = NULL;
-      generate_error (&dtp->common, LIBERROR_END, NULL);
-      return;
-    }
-
   /* Look for &namelist_name .  Skip all characters, testing for $nmlname.
      Exit on success or EOF. If '?' or '=?' encountered in stdin, print
      node names or namelist on stdout.  */
 
 find_nml_name:
-  switch (c = next_char (dtp))
+  c = next_char (dtp);
+  switch (c)
     {
     case '$':
     case '&':
@@ -2903,6 +3016,10 @@ find_nml_name:
 
     case '?':
       nml_query (dtp, '?');
+      goto find_nml_name;
+
+    case EOF:
+      return;
 
     default:
       goto find_nml_name;
@@ -2934,32 +3051,25 @@ find_nml_name:
       if (nml_get_obj_data (dtp, &prev_nl, nml_err_msg, sizeof nml_err_msg)
 			    == FAILURE)
 	{
-	  gfc_unit *u;
-
 	  if (dtp->u.p.current_unit->unit_number != options.stdin_unit)
 	    goto nml_err_ret;
-
-	  u = find_unit (options.stderr_unit);
-	  st_printf ("%s\n", nml_err_msg);
-	  if (u != NULL)
-	    {
-	      sflush (u->s);
-	      unlock_unit (u);
-	    }
+	  generate_error (&dtp->common, LIBERROR_READ_VALUE, nml_err_msg);
         }
 
-   }
+      /* Reset the previous namelist pointer if we know we are not going
+	 to be doing multiple reads within a single namelist object.  */
+      if (prev_nl && prev_nl->var_rank == 0)
+	prev_nl = NULL;
+    }
 
-  dtp->u.p.eof_jump = NULL;
   free_saved (dtp);
   free_line (dtp);
   return;
 
-  /* All namelist error calls return from here */
 
 nml_err_ret:
 
-  dtp->u.p.eof_jump = NULL;
+  /* All namelist error calls return from here */
   free_saved (dtp);
   free_line (dtp);
   generate_error (&dtp->common, LIBERROR_READ_VALUE, nml_err_msg);
